/*****************************************************************************/
/* Object which manages a single USB xb2dev and handles all the IO           */
/* transactions for that device by assigning blkback2_resources   */
/* to process them.                                                          */
/*                                                                           */
/* Copyright (c) 2005 Harry Butterworth IBM Corporation                      */
/*                                                                           */
/* This program is free software; you can redistribute it and/or modify it   */
/* under the terms of the GNU General Public License as published by the     */
/* Free Software Foundation; either version 2 of the License, or (at your    */
/* option) any later version.                                                */
/*                                                                           */
/* This program is distributed in the hope that it will be useful, but       */
/* WITHOUT ANY WARRANTY; without even the implied warranty of                */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General */
/* Public License for more details.                                          */
/*                                                                           */
/* You should have received a copy of the GNU General Public License along   */
/* with this program; if not, write to the Free Software Foundation, Inc.,   */
/* 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                   */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/* Based on arch/xen/drivers/usbif/backend/main.c, original copyright notice */
/* follows...                                                                */
/*****************************************************************************/

/******************************************************************************
 * arch/xen/drivers/usbif/backend/main.c
 * 
 * Backend for the Xen virtual USB driver - provides an abstraction of a
 * USB host controller to the corresponding frontend driver.
 *
 * by Mark Williamson
 * Copyright (c) 2004 Intel Research Cambridge
 * Copyright (c) 2004, 2005 Mark Williamson
 *
 * Based on arch/xen/drivers/blkif/backend/main.c
 * Copyright (c) 2003-2004, Keir Fraser & Steve Hand
 */

#include <linux/blkdev.h>
#include <linux/rwsem.h>
#include <linux/usb.h>
#include <asm-xen/xen-public/io/blkif2.h>
#include "blkback2_assert.h"
#include "blkback2_device.h"
#include "blkback2_trace.h"

typedef enum {
	blkback2_device_stimulus_cn,
	blkback2_device_stimulus_dn,
	blkback2_device_stimulus_rq,
	blkback2_device_stimulus_sd,
	blkback2_device_stimulus_ri,
	blkback2_device_stimulus_rr
} blkback2_device_stimulus;

typedef enum {
	blkback2_device_state_i
} blkback2_device_state;

#define BLKBACK2_RESOURCE_COUNT 4

struct blkback2_device {
	struct xenbus2_device    *xb2dev;
	struct block_device      *blkdev;
	spinlock_t                lock;
	blkback2_device_state     state;
	struct list_head          transaction_list;
	struct list_head          free_resource_list;
	struct blkback2_resource  resources[BLKBACK2_RESOURCE_COUNT];
	struct xenidc_work        open_blkdev_1_work;
	struct xenidc_work        close_blkdev_1_work;
	dev_t                     dev;
	struct xenidc_callback   *endpoint_disconnect_callback;
	unsigned long             current_transactions;
	int                       shutdown;
};

static void
blkback2_device_handle_stimulus(struct blkback2_device *device,
blkback2_device_stimulus stimulus);

static void blkback2_device_open_blkdev_1(void *context);
static void blkback2_device_close_blkdev_1(void *context);

static int
blkback2_device_init_or_exit(struct blkback2_device *device, int exit)
{
	int return_value = 0, i;
	trace();
	if (exit)
		goto exit_path;
	spin_lock_init(&device->lock);
	device->state = blkback2_device_state_i;
	INIT_LIST_HEAD(&device->transaction_list);
	INIT_LIST_HEAD(&device->free_resource_list);
	for (i = 0; i < BLKBACK2_RESOURCE_COUNT; i++) {
		struct blkback2_resource *resource = &device->resources[i];
		return_value = blkback2_resource_init(resource, device);
		if (return_value != 0)
			goto exit_no_resource;
		list_add(&resource->link, &device->free_resource_list);
	}
	xenidc_work_init(&device->open_blkdev_1_work,
				blkback2_device_open_blkdev_1, device);
	xenidc_work_init(&device->close_blkdev_1_work,
				blkback2_device_close_blkdev_1, device);
	return 0;
 exit_path:
 exit_no_resource:
	while (!list_empty(&device->free_resource_list)) {
		struct blkback2_resource *resource = list_entry(
					device->free_resource_list.next,
					struct blkback2_resource, link);
		list_del_init(&resource->link);
		blkback2_resource_exit(resource);
	}
	return return_value;
}

static struct blkback2_device *
blkback2_device_allocate(struct xenbus2_device *xb2dev)
{
	struct blkback2_device *device = (struct blkback2_device *)kmalloc(
				sizeof(struct blkback2_device), GFP_KERNEL);
	if( device != NULL ) {
		memset(device, 0, sizeof(*device));
		device->xb2dev = xb2dev;
		if (blkback2_device_init_or_exit(device, 0) != 0) {
			kfree(device);
			device = NULL;
		}
	}
	return device;
}

static int
blkback2_device_reconfigure(struct blkback2_device *device,
struct xenidc_lbr paramaters, struct xenidc_lbr status)
{
    trace();

/* FIXME */

    return 0;
}

static void blkback2_device_endpoint_connect(struct blkback2_device *device)
{
	unsigned long flags;
	trace();
	spin_lock_irqsave(&device->lock, flags);
	blkback2_device_handle_stimulus(device, blkback2_device_stimulus_cn);
	spin_unlock_irqrestore(&device->lock, flags);
}

static void
blkback2_device_handle_io(struct blkback2_device *device,
struct xenidc_endpoint_transaction *transaction)
{
	unsigned long flags;
	trace();
	spin_lock_irqsave(&device->lock, flags);
	list_add_tail(xenidc_endpoint_transaction_to_link(transaction),
						&device->transaction_list);
	blkback2_device_handle_stimulus(device, blkback2_device_stimulus_rq);
	spin_unlock_irqrestore(&device->lock, flags);
}

static void
blkback2_device_transaction_handler(struct blkback2_device *device,
struct xenidc_endpoint_transaction *transaction)
{
	struct blkif2_transaction_parameters_header header;
	if (xenidc_lbr_copy_out(xenidc_endpoint_transaction_to_parameters_lbr(
				transaction), &header, sizeof(header)) ==
							sizeof(header)) {
		switch (header.transaction_type) {
		case BLKIF2_TRANSACTION_TYPE_IO:
			blkback2_device_handle_io(device, transaction);
			break;
		default:
			xenidc_endpoint_transaction_complete(transaction,
					XENIDC_ERROR_INVALID_PARAMETER);
			break;
		}
	} else {
		/* Parameters were underlength. */
		xenidc_endpoint_transaction_complete(transaction,
					XENIDC_ERROR_INVALID_PROTOCOL);
	}
}

static void
blkback2_device_endpoint_disconnect(struct blkback2_device *device,
struct xenidc_callback *callback)
{
	unsigned long flags;
	trace();
	spin_lock_irqsave(&device->lock, flags);
	device->endpoint_disconnect_callback = callback;
	blkback2_device_handle_stimulus(device, blkback2_device_stimulus_dn);
	spin_unlock_irqrestore(&device->lock, flags);
}

static void blkback2_device_shutdown(struct blkback2_device *device)
{
	trace();
	spin_lock_irq(&device->lock);
	device->shutdown = 0;
	blkback2_device_handle_stimulus(device, blkback2_device_stimulus_sd);
	spin_unlock_irq(&device->lock);
	xenidc_work_until(device->shutdown);
	flush_scheduled_work();
	(void)blkback2_device_init_or_exit(device, 1);
	kfree(device);
}

static void
blkback2_device_invalid_stimulus(struct blkback2_device *device,
blkback2_device_stimulus stimulus)
{
	trace();
	printk(KERN_ERR "blkback2: device %p in state %d"
	     "received invalid stimulus %d", device, device->state, stimulus);
}

static void blkback2_device_open_blkdev(struct blkback2_device *device)
{
	trace();
	xenidc_work_schedule(&device->open_blkdev_1_work);
}

static void blkback2_device_open_blkdev_1(void *context)
{
	struct blkback2_device *device = context;
	trace();
	device->blkdev = open_by_devnum(device->dev, FMODE_READ|FMODE_WRITE);
	if (IS_ERR(device->blkdev))
		goto open_error;
	if (device->blkdev->bd_disk == NULL)
		goto no_disk;
	return;
 no_disk:
	blkdev_put(device->blkdev);
 open_error:
	return;
}

static void blkback2_device_close_blkdev(struct blkback2_device *device)
{
	trace();
	xenidc_work_schedule(&device->close_blkdev_1_work);
}

static void blkback2_device_close_blkdev_1(void *context)
{
	struct blkback2_device *device = context;
	trace();
	blkdev_put(device->blkdev);
}

static void blkback2_device_purge_io(struct blkback2_device *device)
{
	int i;
	trace();
	while (!list_empty(&device->transaction_list)) {
		struct xenidc_endpoint_transaction *transaction = list_entry(
				device->transaction_list.next,
				struct xenidc_endpoint_transaction,
				XENIDC_ENDPOINT_TRANSACTION_LINK);
		list_del_init(xenidc_endpoint_transaction_to_link(
						transaction));
		xenidc_endpoint_transaction_complete(transaction,
					XENIDC_ERROR_DISCONNECT);
	}
	for (i = 0; i < BLKBACK2_RESOURCE_COUNT; i++) {
		struct blkback2_resource *resource = &device->resources[i];
		blkback2_resource_flush_transaction(resource);
	}
}

static void blkback2_device_test_io(struct blkback2_device *device)
{
	trace();
	if (device->current_transactions == 0) {
		blkback2_device_handle_stimulus(device,
						blkback2_device_stimulus_ri);
	} else {
		blkback2_device_handle_stimulus(device,
						blkback2_device_stimulus_rr);
	}
}

static void blkback2_device_kick_io(struct blkback2_device *device)
{
	trace();
	while (!list_empty(&device->transaction_list) &&
				!list_empty(&device->free_resource_list)) {
		struct xenidc_endpoint_transaction *transaction = list_entry(
					device->transaction_list.next,
					struct xenidc_endpoint_transaction,
					XENIDC_ENDPOINT_TRANSACTION_LINK);
		struct blkback2_resource *resource = list_entry(
					device->free_resource_list.next,
					struct blkback2_resource,
					link);
		list_del_init(xenidc_endpoint_transaction_to_link(
								transaction));
		list_del_init(&resource->link);
		device->current_transactions++;
		blkback2_resource_start_transaction(resource, transaction);
	}
}

void
blkback2_device_resource_completed_io(struct blkback2_device *device,
struct blkback2_resource *resource)
{
	unsigned long flags;
	trace();
	spin_lock_irqsave(&device->lock, flags);
	list_add(&resource->link, &device->free_resource_list);
	device->current_transactions--;
	blkback2_device_handle_stimulus(device, blkback2_device_stimulus_rq);
	spin_unlock_irqrestore(&device->lock, flags);
}

static void
blkback2_device_complete_endpoint_disconnect(struct blkback2_device *device)
{
	trace();
	xenidc_callback_success(device->endpoint_disconnect_callback);
}

static void blkback2_device_complete_shutdown(struct blkback2_device *device)
{
	trace();
	device->shutdown = 1;
	xenidc_work_wake_up();
}

struct xenbus2_device *
blkback2_device_query_xb2dev(struct blkback2_device *device)
{
	trace();
	return device->xb2dev;
}

struct block_device *
blkback2_device_query_blkdev(struct blkback2_device *device)
{
	trace();
	return device->blkdev;
}

static void
blkback2_device_handle_stimulus(struct blkback2_device *device,
blkback2_device_stimulus stimulus)
{
	trace_info("device %p in state %d received stimulus %d", device,
						device->state, stimulus);
	switch (device->state) {
	/*FIXME: implement state machine*/
	case blkback2_device_state_i:
		switch (stimulus) {
		case blkback2_device_stimulus_cn:
			break;
		default:
			blkback2_device_invalid_stimulus(device, stimulus);
			break;
		}
		break;
	default:
		blkback2_device_invalid_stimulus(device, stimulus);
		break;
	}
}

static struct xenbus2_driver blkback2_device_class_xenbus2_driver = {
	.allocate    = blkback2_device_allocate,
	.reconfigure = blkback2_device_reconfigure,
	.connect     = blkback2_device_endpoint_connect,
	.transaction = blkback2_device_transaction_handler,
	.disconnect  = blkback2_device_endpoint_disconnect,
	.free        = blkback2_device_shutdown
};

static int blkback2_device_class_init_or_exit(int exit)
{
	int return_value = 0;
	trace();
	if (exit)
		goto exit_path;
	return_value = blkback2_resource_class_init();
	if (return_value != 0)
		goto exit_no_resource_class;
	return_value = xenbus2_register(&blkback2_device_class_xenbus2_driver);
	if (return_value != 0)
		goto exit_no_xenbus2_register;
	return 0;
 exit_path:
	xenbus2_deregister(&blkback2_device_class_xenbus2_driver);
 exit_no_xenbus2_register:
	blkback2_resource_class_exit();
 exit_no_resource_class:
	return return_value;
}

int blkback2_device_class_init(void)
{
	trace();
	return blkback2_device_class_init_or_exit(0);
}

void blkback2_device_class_exit(void)
{
	trace();
	(void)blkback2_device_class_init_or_exit(1);
}

module_init(blkback2_device_class_init);
module_exit(blkback2_device_class_exit);

MODULE_LICENSE("GPL");
