/*****************************************************************************/
/* Copyright (c) 2005 Harry Butterworth IBM Corporation                      */
/*                                                                           */
/* This program is free software; you can redistribute it and/or modify it   */
/* under the terms of the GNU General Public License as published by the     */
/* Free Software Foundation; either version 2 of the License, or (at your    */
/* option) any later version.                                                */
/*                                                                           */
/* This program is distributed in the hope that it will be useful, but       */
/* WITHOUT ANY WARRANTY; without even the implied warranty of                */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General */
/* Public License for more details.                                          */
/*                                                                           */
/* You should have received a copy of the GNU General Public License along   */
/* with this program; if not, write to the Free Software Foundation, Inc.,   */
/* 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                   */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/* Based on:                                                                 */
/*                                                                           */
/* drivers/xen/blkback/blkback.c                                             */
/* arch/xen/drivers/usbif/backend/main.c,                                    */
/*                                                                           */
/* original copyright notices follow...                                      */
/*                                                                           */
/*****************************************************************************/

/******************************************************************************
 * arch/xen/drivers/usbif/backend/main.c
 * 
 * Backend for the Xen virtual USB driver - provides an abstraction of a
 * USB host controller to the corresponding frontend driver.
 *
 * by Mark Williamson
 * Copyright (c) 2004 Intel Research Cambridge
 * Copyright (c) 2004, 2005 Mark Williamson
 *
 * Based on arch/xen/drivers/blkif/backend/main.c
 * Copyright (c) 2003-2004, Keir Fraser & Steve Hand
 */

/******************************************************************************
 * arch/xen/drivers/blkif/backend/main.c
 * 
 * Back-end of the driver for virtual block devices. This portion of the
 * driver exports a 'unified' block-device interface that can be accessed
 * by any operating system that implements a compatible front end. A 
 * reference front-end implementation can be found in:
 *  arch/xen/drivers/blkif/frontend
 * 
 * Copyright (c) 2003-2004, Keir Fraser & Steve Hand
 * Copyright (c) 2005, Christopher Clark
 */

#include <linux/bio.h>
#include <asm-xen/xen-public/io/blkif2.h>
#include "blkback2_assert.h"
#include "blkback2_device.h"
#include "blkback2_resource.h"
#include "blkback2_trace.h"

static struct xenidc_rbr_mapper_pool *rbr_mapper_pool;

int blkback2_resource_class_init(void)
{
	struct xenidc_buffer_resource_list sum, bulk_list;
	trace();
	sum = xenidc_buffer_resource_list_null();
	/* FIXME: not necessarily grant-table RBRs */
	bulk_list = xenidc_grant_table_calculate_buffer_resource_list(
		BLKIF2_MAX_BULK_BYTE_COUNT, BLKIF2_BULK_BYTE_ALIGNMENT);
	xenidc_buffer_resource_list_plus_equals(&sum, &bulk_list);
	rbr_mapper_pool = xenidc_allocate_rbr_mapper_pool(sum);
	return (rbr_mapper_pool != NULL) ? 0 : -ENOMEM;
}

void blkback2_resource_class_exit(void)
{
	trace();
	xenidc_free_rbr_mapper_pool(rbr_mapper_pool);
}

typedef enum {
	blkback2_resource_stimulus_st,	/* Start IO          */
	blkback2_resource_stimulus_ft,	/* Flush transaction */
	blkback2_resource_stimulus_ms,	/* Map success       */
	blkback2_resource_stimulus_mf,	/* Map failure       */
	blkback2_resource_stimulus_bc 	/* BIO completed     */
} blkback2_resource_stimulus;

static void
blkback2_resource_handle_stimulus(struct blkback2_resource *resource,
blkback2_resource_stimulus stimulus);

static void blkback2_resource_map_callback(struct xenidc_callback *callback);
static void blkback2_resource_unmap_callback(struct xenidc_callback *callback);
static void blkback2_resource_submit_bio_1(void *context);

int
blkback2_resource_init(struct blkback2_resource *resource,
struct blkback2_device *device)
{
	trace();
	resource->device = device;
	INIT_LIST_HEAD(&resource->link);
	spin_lock_init(&resource->lock);
	resource->state = blkback2_resource_state_i;
	xenidc_reserve_and_map_rbr_request_init(
				&resource->reserve_and_map_rbr_request,
				blkback2_resource_map_callback,
				blkback2_resource_unmap_callback);
	xenidc_map_rbr_request_element_init(&resource->request_element);
	xenidc_reserve_and_map_rbr_request_add_element(
				&resource->reserve_and_map_rbr_request,
				&resource->request_element);
	xenidc_work_init(&resource->submit_bio_1_work,
				blkback2_resource_submit_bio_1, resource);
	return 0;
}

void blkback2_resource_exit(struct blkback2_resource *resource)
{
	trace();
}

void
blkback2_resource_start_transaction(struct blkback2_resource *resource,
struct xenidc_endpoint_transaction *transaction)
{
	unsigned long flags;
	trace();
	spin_lock_irqsave(&resource->lock, flags);
	resource->transaction       = transaction;
	resource->transaction_error = XENIDC_ERROR_SUCCESS;
	resource->rbr_mapped        = 0;
	blkback2_resource_handle_stimulus(resource,
						blkback2_resource_stimulus_st);
	spin_unlock_irqrestore(&resource->lock, flags);
}

void blkback2_resource_flush_transaction(struct blkback2_resource *resource)
{
	unsigned long flags;
	trace();
	spin_lock_irqsave(&resource->lock, flags);
	blkback2_resource_handle_stimulus(resource,
						blkback2_resource_stimulus_ft);
	spin_unlock_irqrestore(&resource->lock, flags);
}

static void
blkback2_resource_invalid_stimulus(struct blkback2_resource *resource,
blkback2_resource_stimulus stimulus);

static void blkback2_resource_map_buffer(struct blkback2_resource *resource);
static void blkback2_resource_abort_map(struct blkback2_resource *resource);
static void blkback2_resource_submit_bio(struct blkback2_resource *resource);
static void blkback2_resource_set_error(struct blkback2_resource *resource);
static void blkback2_resource_complete(struct blkback2_resource *resource);

static void
blkback2_resource_handle_stimulus(struct blkback2_resource *resource,
blkback2_resource_stimulus stimulus)
{
	trace_info("device resource %p in state %d received stimulus %d",
					resource, resource->state, stimulus);
	switch (resource->state) {
	case blkback2_resource_state_i:
		switch (stimulus) {
		case blkback2_resource_stimulus_st:
			resource->state = blkback2_resource_state_i_st;
			blkback2_resource_map_buffer(resource);
			break;
		case blkback2_resource_stimulus_ft:
			break;
		default:
			blkback2_resource_invalid_stimulus(resource, stimulus);
			break;
		}
		break;
	case blkback2_resource_state_i_st:
		switch (stimulus) {
		case blkback2_resource_stimulus_ft:
			resource->state = blkback2_resource_state_i_st_ft;
			blkback2_resource_abort_map(resource);
			break;
		case blkback2_resource_stimulus_ms:
			resource->state = blkback2_resource_state_i_st_ms;
			blkback2_resource_submit_bio(resource);
			break;
		case blkback2_resource_stimulus_mf:
			resource->state = blkback2_resource_state_i;
			blkback2_resource_complete(resource);
			break;
		default:
			blkback2_resource_invalid_stimulus(resource, stimulus);
			break;
		}
		break;
	case blkback2_resource_state_i_st_ft:
		switch (stimulus) {
		case blkback2_resource_stimulus_ft:
			break;
		case blkback2_resource_stimulus_ms:
			resource->state = blkback2_resource_state_i;
			blkback2_resource_set_error(resource);
			blkback2_resource_complete(resource);
			break;
		case blkback2_resource_stimulus_mf:
			resource->state = blkback2_resource_state_i;
			blkback2_resource_complete(resource);
			break;
		default:
			blkback2_resource_invalid_stimulus(resource, stimulus);
			break;
		}
		break;
	case blkback2_resource_state_i_st_ms:
		switch (stimulus) {
		case blkback2_resource_stimulus_ft:
			break;
		case blkback2_resource_stimulus_bc:
			resource->state = blkback2_resource_state_i;
			blkback2_resource_complete(resource);
			break;
		default:
			blkback2_resource_invalid_stimulus(resource, stimulus);
			break;
		}
		break;
	default:
		blkback2_resource_invalid_stimulus(resource, stimulus);
		break;
	}
}

static void
blkback2_resource_invalid_stimulus(struct blkback2_resource *resource,
blkback2_resource_stimulus stimulus)
{
	trace();
	printk(KERN_ERR "blkback2: device resource %p in state %d received "
		"invalid stimulus %d", resource, resource->state, stimulus);
}

static void
blkback2_resource_map_buffer(struct blkback2_resource *resource)
{
	/*FIXME parameter underlength checked anywhere?*/
	struct blkif2_io_transaction_parameters *parameters = xenidc_vaddress(
				xenidc_endpoint_transaction_to_parameters_lbr(
						resource->transaction));
	trace();
	xenidc_map_rbr_request_element_set_rbr(&resource->request_element,
		     parameters->rbr,
		     (parameters->io_type == BLKIF2_IO_TYPE_READ) ?
		     XENIDC_MAP_RBR_REQUEST_ELEMENT_ACCESS_FLAGS_WRITE :
		     XENIDC_MAP_RBR_REQUEST_ELEMENT_ACCESS_FLAGS_READ);
	xenidc_rbr_mapper_pool_reserve_and_map_rbrs(rbr_mapper_pool,
		&resource->reserve_and_map_rbr_request,
		xenbus2_device_query_address(blkback2_device_query_xb2dev(
							resource->device)));
	return;
}

static void
blkback2_resource_abort_map(struct blkback2_resource *resource)
{
	trace();
	xenidc_rbr_mapper_pool_abort_reserve_and_map_rbrs(
				&resource->reserve_and_map_rbr_request,
				XENIDC_ERROR_DISCONNECT);
}

static void
blkback2_resource_map_callback(struct xenidc_callback *callback)
{
	struct xenidc_reserve_and_map_rbr_request *request =
		xenidc_reserve_and_map_rbr_request_map_callback_to(callback);
	struct blkback2_resource *resource = container_of(request,
			struct blkback2_resource,
			reserve_and_map_rbr_request);
	unsigned long flags;
	trace();
	spin_lock_irqsave(&resource->lock, flags);
	resource->transaction_error = xenidc_callback_query_error(callback);
	if (resource->transaction_error == XENIDC_ERROR_SUCCESS) {
		resource->rbr_mapped = 1;
		blkback2_resource_handle_stimulus(resource,
				blkback2_resource_stimulus_ms);
	} else {
		trace_info("failed to map FE buffer");
		blkback2_resource_handle_stimulus(resource,
				blkback2_resource_stimulus_mf);
	}
	spin_unlock_irqrestore(&resource->lock, flags);
}

static int
blkback2_resource_submit_bio_2(struct bio *bio, unsigned int done, int error);
static void
blkback2_resource_submit_bio_3(struct blkback2_resource *resource);

static void blkback2_resource_submit_bio(struct blkback2_resource *resource)
{
	trace();
	xenidc_work_schedule(&resource->submit_bio_1_work);
}

static void blkback2_resource_submit_bio_1(void *context)
{
	struct blkback2_resource *resource = context;
	struct blkif2_io_transaction_parameters *parameters = xenidc_vaddress(
				xenidc_endpoint_transaction_to_parameters_lbr(
						resource->transaction));
	int operation = (parameters->io_type == BLKIF2_IO_TYPE_READ) ?
								READ : WRITE;
	struct xenidc_lbr top_lbr = *xenidc_map_rbr_request_element_to_lbr(
			&resource->request_element);
	blkif2_lba lba = parameters->lba;
	struct bio *bio = NULL;
	trace();
	/* FIXME: need to check lba in range, segment alignment and fail any */
	/* writes to ro devices. */
	atomic_set(&resource->completions_out, 1);
	while (xenidc_lbr_query_byte_count(&top_lbr) != 0) {
		struct xenidc_lbr segment_lbr = xenidc_lbr_resolve(&top_lbr);
		xenidc_lbr_advance(&top_lbr,
				xenidc_lbr_query_byte_count(&segment_lbr));
		do {
			xenidc_buffer_byte_count segment_byte_count =
				xenidc_lbr_query_byte_count(&segment_lbr);
			void *segment_address = xenidc_vaddress(&segment_lbr);
			xenidc_buffer_byte_count offset_this_page =
			(unsigned long)segment_address & ~PAGE_MASK;
			xenidc_buffer_byte_count byte_count_this_page = min(
					segment_byte_count,
					(xenidc_buffer_byte_count)PAGE_SIZE -
							offset_this_page);
			xenidc_lbr_advance(&segment_lbr, byte_count_this_page);
			while ((bio == NULL) || (bio_add_page(bio,
					virt_to_page(segment_address),
					byte_count_this_page,
					offset_this_page) == 0)) {
				if (bio != NULL) {
					atomic_inc(&resource->completions_out);
					submit_bio(operation, bio);
				}
				bio = bio_alloc(GFP_KERNEL, 16);/*FIXME:magic*/
				/* FIXME: resource management sucks here. */
				BUG_ON(bio == NULL);
				bio->bi_bdev    = blkback2_device_query_blkdev(
							resource->device);
				bio->bi_private = resource;
				bio->bi_end_io  =
					blkback2_resource_submit_bio_2;
				bio->bi_sector  = lba;
			}
			lba += byte_count_this_page >> 9;
		} while(xenidc_lbr_query_byte_count(&segment_lbr) != 0);
	}
	if (bio != NULL) {
		atomic_inc(&resource->completions_out);
		submit_bio(operation, bio);
	}
	blkback2_resource_submit_bio_3(resource);
}

static int
blkback2_resource_submit_bio_2(struct bio *bio, unsigned int done, int error)
{
	unsigned long flags;
	struct blkback2_resource *resource = bio->bi_private;
	trace();
	if(error) {
		spin_lock_irqsave(&resource->lock, flags);
		if( resource->transaction_error == XENIDC_ERROR_SUCCESS )
			resource->transaction_error =
					blkif2_error_map_local_to(error);
		spin_unlock_irqrestore(&resource->lock, flags);
	}
	if (bio->bi_size != 0)
		return 1;
	if (!test_bit(BIO_UPTODATE, &bio->bi_flags)) {
		spin_lock_irqsave(&resource->lock, flags);
		if( resource->transaction_error == XENIDC_ERROR_SUCCESS )
			resource->transaction_error =
					blkif2_error_map_local_to(-EIO);
		spin_unlock_irqrestore(&resource->lock, flags);
	}
	bio_put(bio);
        blkback2_resource_submit_bio_3(resource);
	return 0;
}

static void blkback2_resource_submit_bio_3(struct blkback2_resource *resource)
{
	trace();
	if (atomic_dec_and_test(&resource->completions_out)) {
		unsigned long flags;
		spin_lock_irqsave(&resource->lock, flags);
		blkback2_resource_handle_stimulus(resource,
						blkback2_resource_stimulus_bc);
		spin_unlock_irqrestore(&resource->lock, flags);
	}
}

static void blkback2_resource_set_error(struct blkback2_resource *resource)
{
	trace();
	resource->transaction_error = XENIDC_ERROR_DISCONNECT;
}

static void blkback2_resource_complete(struct blkback2_resource *resource)
{
	trace();
	if (resource->rbr_mapped) {
		xenidc_rbr_mapper_pool_unmap_and_unreserve_rbrs(
				&resource->reserve_and_map_rbr_request);
	} else {
		xenidc_callback_success(
			xenidc_reserve_and_map_rbr_request_to_unmap_callback(
				&resource->reserve_and_map_rbr_request));
	}
}

static void
blkback2_resource_unmap_callback(struct xenidc_callback *callback)
{
	struct xenidc_reserve_and_map_rbr_request *request =
		xenidc_reserve_and_map_rbr_request_unmap_callback_to(callback);
	struct blkback2_resource *resource = container_of(request,
				struct blkback2_resource,
				reserve_and_map_rbr_request);
	trace();
	xenidc_endpoint_transaction_complete(resource->transaction,
						resource->transaction_error);
	blkback2_device_resource_completed_io(resource->device, resource);
}
